import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_launch_conditions(bs):
    """Searches the launch conditions in the bs object. Returns a dict with the
    settings.

    Parameters
    ----------
    bs : BeautifulSoup
        The BeautifulSoup object of the open rocket file.

    Returns
    -------
    settings : dict
        A dict containing the settings for the launch conditions.
    """
    settings = {}

    launch_rod_length = float(bs.find("launchrodlength").text)
    launch_rod_angle = float(bs.find("launchrodangle").text)
    launch_rod_direction = float(bs.find("launchroddirection").text)
    logger.info(
        "Collected launch conditions: launch rod length, launch rod angle, "
        "launch rod direction."
    )

    settings = {
        "rail_length": launch_rod_length,
        "inclination": 90 - launch_rod_angle,
        "heading": launch_rod_direction,
    }
    logger.info("Exported launch conditions.\n%s", _dict_to_string(settings, indent=23))
    return settings



import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_nosecone(bs, elements=None, rocket_radius=None, just_radius=False):
    """Search for the nosecone in the bs and return the settings as a dict.

    Parameters
    ----------
    bs : bs4.BeautifulSoup
        The BeautifulSoup object of the .ork file.
    elements : dict
        Dictionary with the elements of the rocket.
    rocket_radius : float
        The radius of the rocket.
    just_radius : bool
        If True, returns only the radius of the nosecone.

    Returns
    -------
    settings : dict
        Dictionary with the settings for the nosecone.
    """
    settings = {}
    nosecone = bs.find("nosecone")  # TODO: allow for multiple nosecones
    name = nosecone.find("name").text if nosecone else "nosecone"

    if not nosecone:
        nosecones = list(
            filter(
                lambda x: x.find("name").text == "Nosecone", bs.findAll("transition")
            )
        )
        if len(nosecones) == 0:
            logger.info("Could not fetch a nosecone")
            return settings
        if len(nosecones) > 1:
            logger.info("Multiple nosecones found, using only the first one")
        nosecone = nosecones[0]  # only the first nosecone is considered

    length = float(nosecone.find("length").text)
    kind = nosecone.find("shape").text
    base_radius = nosecone.find("aftradius").text
    try:
        base_radius = float(base_radius)
    except ValueError:
        if "auto" in base_radius:
            # TODO: this is a temporary fix, we should look to the next component radius
            base_radius = rocket_radius

    # TODO: this is for specific use in the get_rocket_radius() function, given that the
    # search_nosecone is changed, the get_rocket_radius function must be re-evaluated
    if just_radius:
        return base_radius  # return nosecone radius to the get_rocket_radius function

    def get_position(name, length):
        count = 0
        position = None
        # defaults to case insensitive
        lower_name = name.lower()
        for element in elements.values():
            if element["name"].lower() == lower_name and element["length"] == length:
                count += 1
                position = element["position"]
        if count > 1:
            logger.warning(
                "Multiple nosecones with the same name and length, "
                "using the last one found."
            )
        elif count == 0:
            logger.error(
                "No element with the name '%s' and length '%f' was found",
                name,
                length,
            )
        return position

    logger.info("Collected the dimensions of the nosecone: length, shape and position.")

    if kind == "haack":
        logger.info("Nosecone is a haack nosecone, searching for the shape parameter")

        shape_parameter = float(nosecone.find("shapeparameter").text)
        kind = "Von Karman" if shape_parameter == 0.0 else "lvhaack"
        settings.update({"noseShapeParameter": shape_parameter})
        logger.info("Shape parameter of the nosecone: %s", shape_parameter)

    settings = {
        "name": name,
        "kind": kind,
        "length": length,
        "base_radius": base_radius,
        "position": get_position(name, length),
    }
    logger.info("Nosecone setting defined:\n %s", _dict_to_string(settings, indent=23))
    return settings



import logging

import numpy as np

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_parachutes(bs):
    """Search for the parachutes in the bs and return the settings as a dict.

    Parameters
    ----------
    bs : bs4.BeautifulSoup
        The BeautifulSoup object of the .ork file.

    Returns
    -------
    settings : dict
        A dict containing the settings for the parachutes. The keys are integers
        and the values are dicts containing the settings for each parachute.
        The keys of the parachute dicts are: "name", "cd", "cds", "area",
        "deploy_event", "deploy_delay", "deploy_altitude".
    """
    settings = {}

    chutes = bs.findAll("parachute")
    logger.info("A total of %d parachutes were detected", len(chutes))

    for idx, chute in enumerate(chutes):
        logger.info("Starting to collect the settings of the parachute number %d", idx)
        name = chute.find("name").text

        # parachute settings
        cd = "auto" if "auto" in chute.find("cd").text else float(chute.find("cd").text)
        cd = search_cd_chute_if_auto(chute) if cd == "auto" else cd
        area = np.pi * float(chute.find("diameter").text) ** 2 / 4
        cds = cd * area
        logger.info("Parachute '%s' has a drag coefficient of %f", name, cd)

        # deployment settings
        deploy_event = chute.find("deployevent").text
        deploy_delay = float(chute.find("deploydelay").text)
        deploy_altitude = (
            float(chute.find("deployaltitude").text)
            if deploy_event == "altitude"
            else None
        )
        logger.info("Parachute '%s' will deploy at %s", name, deploy_event)

        setting = {
            "name": name,
            "cd": cd,
            "cds": cds,
            "area": area,
            "deploy_event": deploy_event,
            "deploy_delay": deploy_delay,
            "deploy_altitude": deploy_altitude,
        }
        settings[idx] = setting

        logger.info(
            "The Parachute number %d had its settings defined:\n%s",
            idx,
            _dict_to_string(setting, indent=23),
        )
    logger.info("All parachutes settings were collected")
    return settings


def search_cd_chute_if_auto(bs):
    # if the parachute cd is st to "auto", then look for the cd in the next tag
    # return float(
    #     next(
    #         filter(lambda x: x.text.replace(".", "").isnumeric(), bs.findAll("cd"))
    #     ).text
    # )

    # TODO: for the future, we need to check if the ork object has a drag coefficient

    # simply return 1.0
    logger.warning(
        "cd auto: the cd is set to 1.0 for parachute %s", bs.find("name").text
    )
    return 1.0



import logging

logger = logging.getLogger(__name__)


def search_rail_buttons(bs, elements: dict) -> dict:

    lugs_elements = []
    for value in elements.values():
        if value["type"] in ["LaunchLug", "RailButton"]:
            lugs_elements.append(value)

    if len(lugs_elements) < 2 or not lugs_elements:
        logger.info("Could not fetch a pair of rail buttons")
        return {}

    lugs_elements.sort(key=lambda x: x["position"])
    # We only need the 2 buttons closest to the nozzle.
    lugs_elements = lugs_elements[-2:]

    upper_position = lugs_elements[0]["position"]
    lower_position = lugs_elements[1]["position"]
    name = str(lugs_elements[0]["name"])

    angular_position = 0.0
    lugs = bs.findAll("launchlug")
    for lug in lugs:
        if lug.find("name").text == name:
            angular_position = float(lug.find("radialdirection").text)
            break

    return {
        "name": str(name),
        "upper_position": upper_position,
        "lower_position": lower_position,
        "distance": lower_position - upper_position,
        "angular_position": angular_position,
    }



import logging

import numpy as np

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_stored_results(bs, datapoints, data_labels, time_vector, burnout_position):
    """Search for the stored simulation results in the bs and return the
    settings as a dict.

    Parameters
    ----------
    bs : BeautifulSoup
        BeautifulSoup object of the .ork file.

    Returns
    -------
    settings : dict
        A dict containing the settings for the launch conditions. The keys are:
        "maxaltitude", "maxvelocity", "maxacceleration", "maxmach", "timetoapogee",
        "flighttime", "groundhitvelocity" and "launchrodvelocity".
    """
    settings = {}
    sim = bs.find("simulation")
    sim_data = sim.find("flightdata")
    logger.info("Found the 'flightdata' tag in the 'simulation' tag.")
    name_map = {
        "max_altitude": "maxaltitude",
        "max_velocity": "maxvelocity",
        "max_acceleration": "maxacceleration",
        "max_mach": "maxmach",
        "time_to_apogee": "timetoapogee",
        "flight_time": "flighttime",
        "ground_hit_velocity": "groundhitvelocity",
        "launch_rod_velocity": "launchrodvelocity",
    }

    for key, value in name_map.items():
        settings[key] = float(sim_data.get(value, 0))
        logger.info(
            "Retrieved the '%s' value from the .ork file: %s", key, settings[key]
        )

    settings["max_stability_margin"] = __get_parameter(
        datapoints,
        data_labels,
        time_vector,
        "Stability margin calibers",
        position="max",
    )
    settings["min_stability_margin"] = __get_parameter(
        datapoints,
        data_labels,
        time_vector,
        "Stability margin calibers",
        position="min",
    )
    settings["burnout_stability_margin"] = __get_parameter(
        datapoints,
        data_labels,
        time_vector,
        "Stability margin calibers",
        position=burnout_position,
    )
    settings["max_thrust"] = __get_parameter(
        datapoints, data_labels, time_vector, "Thrust", position="max"
    )

    logger.info(
        "The flight data was successfully retrieved:\n%s",
        _dict_to_string(settings, indent=23),
    )
    return settings


def __get_parameter(datapoints, data_labels, time_vector, label, position):
    """Get the latitude and longitude from the .ork file.
    Parameters
    ----------
    label : str
        Latitude or longitude.
    datapoints : list
        List of datapoints from the .ork file.
    data_labels : list
        List of labels for the datapoints.
    time_vector : list
        The time vector of the simulation.
    position : str or int
        The position to get the value from. Can be "last", "first", "max", "min"
        or an integer.
    """

    parameter = [
        float(datapoint.text.split(",")[data_labels.index(label)])
        for datapoint in datapoints
    ]
    # convert to numpy array
    parameter = np.array([time_vector, parameter]).T
    # sort by time
    parameter = parameter[parameter[:, 0].argsort()]
    # clip the curve to remove negative values
    parameter[parameter[:, 1] < 0, 1] = 0
    # Assuming parameter is a NumPy array and 'NaN' values are represented as np.nan
    # This will keep rows where the second column is not NaN
    parameter = parameter[~np.isnan(parameter[:, 1])]

    if isinstance(position, str):
        if position == "last":
            # return the end point (final time, final value)
            return parameter[-1, 1]
        elif position == "first":
            # return the first point (initial time, initial value)
            return parameter[0, 1]
        elif position == "max":
            return np.max(parameter[:, 1])  # return the maximum value
        elif position == "min":
            return np.min(parameter[:, 1])  # return the minimum value
    else:
        pass
    if isinstance(position, np.int64):
        return parameter[position, 1]  # return the value at the specified position
    else:
        logger.error("Invalid position parameter")
        raise ValueError("Error in position parameter")



import logging
import os
import shutil
from pathlib import Path
from zipfile import BadZipFile, ZipFile

from bs4 import BeautifulSoup

logger = logging.getLogger(__name__)


def extract_ork_from_zip(zip_path: Path, extract_dir: Path) -> Path:
    """Extracts .ork file from a zip archive and returns its path.
    This is important because sometimes the .xml file is stuck inside the .ork
    file. The function extracts the .ork (.xml) file from the zip (.ork) file
    and returns its path as a Path object.

    Notes
    -----
    For illustration: if you try to open the .ork file with a text editor and
    you see a bunch of weird characters, it is probably a "zip" file. You can
    try to rename the file to .zip and open it with a zip extractor.

    Parameters
    ----------
    zip_path : Path
        The path to the .zip file.
    extract_dir : Path
        The path to the directory where the .ork file will be extracted.

    Returns
    -------
    Path
        The path to the extracted .ork file.
    """
    try:
        temp_zip_path = zip_path.with_suffix(".zip")
        shutil.copy(zip_path, temp_zip_path)
        with ZipFile(temp_zip_path) as zf:
            zf.extract("rocket.ork", path=extract_dir)
        logger.info(
            'Successfully extracted rocket.ork from "%s" to "%s"',
            zip_path.as_posix(),
            extract_dir.as_posix(),
        )
        return extract_dir / "rocket.ork"
    except BadZipFile:
        logger.warning(
            'The file "%s" seems to be a rocket.ork file and not a compressed archive.',
            zip_path.as_posix(),
        )
        return zip_path
    except Exception as e:
        logger.error(
            'Error while extracting data from "%s": %s', zip_path.as_posix(), e
        )
        raise
    finally:
        if temp_zip_path.exists():
            os.remove(temp_zip_path)


def parse_ork_file(ork_path: Path):
    """Parses the .ork file and returns BeautifulSoup and a list of datapoints.

    Parameters
    ----------
    ork_path : Path
        The path to the .ork file.

    Returns
    -------
    BeautifulSoup
        The BeautifulSoup object.
    """
    try:
        with open(ork_path, encoding="utf-8") as file:
            bs = BeautifulSoup(file, features="xml")
            datapoints = bs.findAll("datapoint")
            logger.info(
                "Successfully parsed .ork file at '%s' with %d datapoints",
                ork_path.as_posix(),
                len(datapoints),
            )
            return bs, datapoints
    except UnicodeDecodeError as exc:
        error_msg = (
            "The .ork file is not in UTF-8."
            + "Please open the .ork file in a text editor and save it as UTF-8."
        )
        logger.error(error_msg)
        raise UnicodeDecodeError(error_msg) from exc
    except Exception as e:
        logger.error("Error while parsing the file '%s': %s", ork_path.as_posix(), e)
        raise e


def _dict_to_string(dictionary, indent=0):
    """Converts a dictionary to a string.

    Parameters
    ----------
    dictionary : dict
        Dictionary to be converted.
    indent : int, optional
        Indentation level, by default 0.

    Returns
    -------
    str
        String representation of the dictionary.

    Examples
    --------
    >>> from rocketserializer._helpers import _dict_to_string
    >>> _dict_to_string({"a": 1, "b": {"c": 2}})
    " a: 1\n b: \n     c: 2\n"
    """
    string = ""
    for key, value in dictionary.items():
        string += " " * indent + str(key) + ": "
        if isinstance(value, dict):
            string += "\n" + _dict_to_string(value, indent + 4)
        else:
            string += str(value) + "\n"
    return string


# if __name__ == "__main__":
#     import doctest

#     doctest.testmod()



import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_rocket(bs, datapoints, data_labels, burnout_position):
    settings = {}

    # get radius
    settings["radius"] = get_rocket_radius(bs)
    logger.info("Collected rocket radius.")

    # get mass
    cg_location_vector = [
        float(datapoint.text.split(",")[data_labels.index("CG location")])
        for datapoint in datapoints
    ]
    settings["mass"] = get_mass(datapoints, data_labels, burnout_position)
    logger.info("Collected rocket mass.")

    # get inertias
    inertia_z, inertia_i = get_inertias(data_labels, burnout_position, datapoints)
    settings["inertia"] = (inertia_i, inertia_i, inertia_z)
    logger.info("Collected rocket inertia.")

    # get center of mass
    center_of_dry_mass = cg_location_vector[burnout_position]
    center_of_mass = cg_location_vector[0]
    rocket_dry_mass = settings["mass"]
    propellant_mass = get_mass(datapoints, data_labels, 0) - rocket_dry_mass

    center_of_propellant_mass = (
        center_of_mass * (rocket_dry_mass + propellant_mass)
        - rocket_dry_mass * center_of_dry_mass
    ) / propellant_mass
    motor_position = center_of_propellant_mass
    settings["center_of_mass_without_propellant"] = center_of_dry_mass
    logger.info("Collected rocket center of mass.")

    # get coordinate system orientation
    settings["coordinate_system_orientation"] = "nose_to_tail"

    logger.info(
        "All the Rocket information was collected:\n%s",
        _dict_to_string(settings, indent=23),
    )
    return settings, motor_position


def get_rocket_radius(bs):
    # We want to take the maximum radius of the rocket
    tubes = bs.findAll("bodytube")
    noses = bs.findAll("nosecone")

    tubes_radius = [i.find("radius").text for i in tubes]
    noses_radius = [i.find("aftradius").text for i in noses]

    all_radius = tubes_radius + noses_radius

    # We need to convert to float, but removing the "auto" string first
    all_radius = [i.replace("auto ", "") for i in all_radius]
    all_radius = [i for i in all_radius if i != "auto"]
    all_radius = [float(i) for i in all_radius]

    rocket_radius = max(all_radius)
    logger.info("The maximum radius of the rocket is: %f", rocket_radius)
    return rocket_radius


def get_mass(datapoints, data_labels, burnout_position):
    mass_vector = [
        float(datapoint.text.split(",")[data_labels.index("Mass")])
        for datapoint in datapoints
    ]
    return mass_vector[burnout_position]


def get_inertias(data_labels, burnout_position, datapoints):
    """Get the moment of inertia of the rocket in the longitudinal and rotational
    axis. The moment of inertia is calculated at the burnout position. This
    means that the motor is included in the calculation, but the propellant mass
    is not.

    Parameters
    ----------
    data_labels : list
        List of strings with the labels of the data.
    burnout_position : int
        The index of the burnout position in the data.
    datapoints : list
        List of datapoints available in the .ork file.

    Returns
    -------
    (longitudinal, rotational) : tuple of floats
        The moment of inertia of the rocket in the longitudinal and rotational
        axis, respectively.
    """
    longitudinal = [
        float(
            datapoint.text.split(",")[
                data_labels.index("Longitudinal moment of inertia")
            ]
        )
        for datapoint in datapoints
    ][burnout_position]
    rotational = [
        float(
            datapoint.text.split(",")[data_labels.index("Rotational moment of inertia")]
        )
        for datapoint in datapoints
    ][burnout_position]
    logger.info(
        "The moment of inertia of the rocket is: %f (longitudinal) and %f (rotational)",
        longitudinal,
        rotational,
    )
    return longitudinal, rotational



import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_transitions(bs, elements, ork):
    """Search for the transitions in the bs and return the settings as a dict.

    Parameters
    ----------
    bs : bs4.BeautifulSoup
        The BeautifulSoup object of the .ork file.
    elements : dict
        Dictionary with the elements of the rocket.
    ork : orhelper
        orhelper object of the open rocket file.

    Returns
    -------
    settings : dict
        Dictionary with the settings for the transitions. The keys are integers
        and the values are dicts containing the settings for each transition.
        The keys of the transition dicts are: "name", "top_radius",
        "bottom_radius", "length", "position".
    """
    settings = {}
    transitions = bs.findAll("transition")
    logger.info("A total of %d transitions were found", len(transitions))

    transitions_ork = [
        ele
        for ele in ork.getRocket().getChild(0).getChildren()
        if ele.getClass().getSimpleName() == "Transition"
    ]  # TODO: only works for a single stage rocket.

    for idx, transition in enumerate(transitions):
        logger.info("Starting to collect the settings of the transition number %d", idx)

        label = transition.find("name").text
        logger.info("Collected the name of the transition number %d", idx)

        transition_ork = transitions_ork[idx]
        top_radius = float(transition_ork.getForeRadius())
        bottom_radius = (
            transition.find("aftradius").text
            if "auto" in transition.find("aftradius").text
            else float(transition.find("aftradius").text)
        )
        length = float(transition.find("length").text)
        logger.info("Collected the dimensions of the transition number %d", idx)

        def get_position(name, length):
            count = 0
            lower_name = name.lower()
            position = None
            for element in elements.values():
                if (
                    element["name"].lower() == lower_name
                    and element["length"] == length
                ):
                    count += 1
                    position = element["position"]
            if count > 1:
                logger.warning(
                    "Multiple transitions with the same name and length, "
                    "using the last one found."
                )
            elif count == 0:
                logger.error(
                    "No element with the name %s and length %f was found",
                    name,
                    length,
                )
            return position

        transition_setting = {
            "name": label,
            "top_radius": top_radius,
            "bottom_radius": bottom_radius,
            "length": length,
            "position": get_position(label, length),
        }
        settings[idx] = transition_setting
        logger.info(
            "The transition number %d was defined with the following settings:\n%s",
            idx,
            _dict_to_string(transition_setting, indent=23),
        )

    logger.info("All the %d transition settings were defined", len(transitions))
    return settings



import json
import logging
import os

import nbformat as nbf

logger = logging.getLogger(__name__)


class NotebookBuilder:
    """Class that takes a parameters.json file and creates a jupyter notebook
    using rocketpy simulation on it
    """

    def __init__(self, parameters_json: str) -> None:
        """read the file and process the dictionary do not build anything yet"""
        self.parameters_json = parameters_json
        self.trapezoidal_fins_check = False
        self.elliptical_fins_check = False
        self.__extract_output_folder_from_parameters_json()
        self.read()
        self.process()

    def __extract_output_folder_from_parameters_json(self):
        self.__output_folder = os.path.dirname(self.parameters_json)

    def read(self) -> dict:
        # read the json file and return the dict and save it to self.parameters dict
        # if any problem happens here, already tell the user
        with open(self.parameters_json, "r", encoding="utf-8") as f:
            self.parameters = json.load(f)
        return self.parameters

    def process(self):
        # TODO: read the dict and search for any inconsistencies
        return self.parameters

    def build(self, destination: str):
        if os.path.isdir(destination):
            self.__output_folder = destination
        else:
            raise FileNotFoundError(
                f"Destination folder '{destination}' not found. Please create it "
                "first or verify if it is really a folder."
            )

        nb = nbf.v4.new_notebook()
        nb = self.build_header(nb)
        nb = self.build_imports(nb)
        nb = self.build_environment(nb)
        nb = self.build_motor(nb)
        nb = self.build_rocket(nb)
        nb = self.build_flight(nb)
        nb = self.build_compare_results(nb)
        self.save_notebook(nb, destination)
        logger.info(
            "[NOTEBOOK BUILDER] Notebook successfully built! You can find it at: %s",
            destination,
        )

    def build_header(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # create the first commentary cell
        text = "# RocketPy Simulation\n"
        text += "This notebook was generated using Rocket-Serializer, a RocketPy"
        text += " tool to convert simulation files to RocketPy simulations\n"
        text += (
            "The notebook was generated using the following parameters file: "
            + f"`{self.parameters_json}`\n"
        )

        nb["cells"] = [nbf.v4.new_markdown_cell(text)]
        logger.info("[NOTEBOOK BUILDER] Header created")
        return nb

    def build_imports(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # install rocketpy
        text = "%pip install rocketpy<=2.0\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # import classes
        text = (
            "from rocketpy import Environment, SolidMotor, Rocket, Flight, "
            + "TrapezoidalFins, EllipticalFins, RailButtons, NoseCone, Tail\n"
        )
        text += "import datetime\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))
        logger.info("[NOTEBOOK BUILDER] Imports section created")
        return nb

    def build_environment(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # add a markdown cell
        text = "## Environment\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        # add a code cell
        text = "env = Environment()\n"
        text += (
            f"env.set_location(latitude={self.parameters['environment']['latitude']}, "
            + f"longitude={self.parameters['environment']['longitude']})\n"
        )
        text += f"env.set_elevation({self.parameters['environment']['elevation']})\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # add a markdown cell
        text = "Optionally, you can set the date and atmospheric model\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        # add a code cell
        text = "tomorrow = datetime.date.today() + datetime.timedelta(days=1)\n"
        text += "env.set_date((tomorrow.year, tomorrow.month, tomorrow.day, 12))\n"
        text += "# env.set_atmospheric_model(type='Forecast', file='GFS')"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # add a code cell
        text = "env.all_info()\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Environment section created")
        return nb

    def build_motor(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # start section and give comments
        text = "## Motor\n"
        text += "Currently, only Solid Motors are supported by Rocket-Serializer. If "
        text += "you want to use a Liquid/Hybrid motor, please use rocketpy directly.\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        thrust_source = self.parameters["motors"]["thrust_source"]
        thrust_source = os.path.relpath(thrust_source, self.__output_folder)

        # define the motor
        text = "motor = SolidMotor(\n"
        text += f"    thrust_source='{thrust_source}',\n"
        text += f"    dry_mass={self.parameters['motors']['dry_mass']},\n"
        text += (
            "    center_of_dry_mass_position="
            + f"{self.parameters['motors']['center_of_dry_mass_position']},\n"
        )
        text += (
            "    dry_inertia="
            + f"{'[' + str(self.parameters['motors']['dry_inertia'])[1:-1] + ']'},\n"
        )
        text += (
            "    grains_center_of_mass_position="
            + f"{self.parameters['motors']['grains_center_of_mass_position']},\n"
        )
        grain_outer_radius = self.parameters["motors"]["grain_outer_radius"]
        grain_initial_height = self.parameters["motors"]["grain_initial_height"]
        text += f"    grain_number={self.parameters['motors']['grain_number']},\n"
        text += f"    grain_density={self.parameters['motors']['grain_density']},\n"
        text += f"    grain_outer_radius={grain_outer_radius},\n"
        text += (
            "    grain_initial_inner_radius="
            + f"{self.parameters['motors']['grain_initial_inner_radius']},\n"
        )
        text += f"    grain_initial_height={grain_initial_height},\n"
        text += (
            f"    grain_separation={self.parameters['motors']['grain_separation']},\n"
        )
        text += f"    nozzle_radius={self.parameters['motors']['nozzle_radius']},\n"
        # text += f"    burn_time={self.parameters['motors']['burn_time']},\n"
        text += f"    nozzle_position={self.parameters['motors']['nozzle_position']},\n"
        text += f"    throat_radius={self.parameters['motors']['throat_radius']},\n"
        text += (
            "    reshape_thrust_curve=False,  # Not implemented in Rocket-Serializer\n"
        )
        text += "    interpolation_method='linear',\n"
        text += (
            "    coordinate_system_orientation="
            + f"'{self.parameters['motors']['coordinate_system_orientation']}',\n"
        )
        text += ")\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # see the outputs
        text = "motor.all_info()\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Motor section created")
        return nb

    def build_rocket(self, nb: nbf.v4.new_notebook):
        # add a markdown cell
        text = "## Rocket\n"
        text += (
            "Currently, only single stage rockets are supported by Rocket-Serializer\n"
        )
        text += (
            "We will start by defining the aerodynamic surfaces, "
            "and then build the rocket.\n"
        )
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        self.build_all_aerodynamic_surfaces(nb)

        drag_curve = self.parameters["rocket"]["drag_curve"]
        drag_curve = os.path.relpath(drag_curve, self.__output_folder)

        # define the Rocket
        inertia_string = "[" + str(self.parameters["rocket"]["inertia"])[1:-1] + "]"
        text = "rocket = Rocket(\n"
        text += f"    radius={self.parameters['rocket']['radius']},\n"
        text += f"    mass={self.parameters['rocket']['mass']},\n"
        text += f"    inertia={inertia_string},\n"
        text += f"    power_off_drag='{drag_curve}',\n"
        text += f"    power_on_drag='{drag_curve}',\n"
        text += (
            "    center_of_mass_without_motor="
            + f"{self.parameters['rocket']['center_of_mass_without_propellant']},\n"
        )
        text += (
            "    coordinate_system_orientation="
            + f"'{self.parameters['rocket']['coordinate_system_orientation']}',\n"
        )
        text += ")\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # add surfaces to rocket
        self.add_surfaces_to_rocket(nb)

        # add the motor to the rocket
        text = (
            "rocket.add_motor(motor, position= "
            + f"{self.parameters['motors']['position']})\n"
        )
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # add a code cell
        text = "### Rocket Info\n"
        text += "rocket.all_info()\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Rocket section created")
        return nb

    def build_all_aerodynamic_surfaces(
        self, nb: nbf.v4.new_notebook
    ) -> nbf.v4.new_notebook:
        """This is simple: receive the current notebook object, start appending
        cells for each aerodynamic surface and return the notebook object"""
        self.build_nosecones(nb)
        self.build_fins(nb)
        self.build_tails(nb)
        self.build_rail_buttons(nb)
        logger.info("[NOTEBOOK BUILDER] All aerodynamic surfaces created.")
        return nb

    def add_surfaces_to_rocket(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # add a markdown cell
        text = "### Adding surfaces to the rocket\n"
        text += "Now that we have all the surfaces, we can add them to the rocket\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))
        text = "rocket.add_surfaces("

        # building surfaces and positions text
        surface_text = "surfaces=["
        position_text = "positions=["

        # adding nosecone
        surface_text += "nosecone, "
        position_text += f"{self.parameters['nosecones']['position']}, "

        # checking and adding fins
        # trapezoidal fins
        if self.trapezoidal_fins_check:
            for i in range(len(self.parameters["trapezoidal_fins"])):
                surface_text += f"trapezoidal_fins[{i}], "
                position_text += (
                    f"{self.parameters['trapezoidal_fins'][str(i)]['position']}, "
                )
        # elliptical fins
        if self.elliptical_fins_check:
            for i in range(len(self.parameters["elliptical_fins"])):
                surface_text += f"elliptical_fins[{i}], "
                position_text += (
                    f"{self.parameters['elliptical_fins'][str(i)]['position']}, "
                )
        # free form fins

        # adding tails
        for i in range(len(self.parameters["tails"])):
            surface_text += f"tails[{i}], "
            position_text += f"{self.parameters['tails'][str(i)]['position']}, "

        # closing surfaces and positions text
        surface_text = surface_text[:-2] + "]"
        position_text = position_text[:-2] + "]"
        text += surface_text + ", " + position_text + ")"
        nb["cells"].append(nbf.v4.new_code_cell(text))
        return nb

    def build_nosecones(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        """Generates a section defining the nosecone and returns the notebook."""
        # add a markdown cell
        text = "### Nosecones\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        # Assumption: only a single nosecone to be added
        text = "nosecone = NoseCone(\n"
        text += f"    length={self.parameters['nosecones']['length']},\n"
        text += f"    kind='{self.parameters['nosecones']['kind']}',\n"
        text += f"    base_radius={self.parameters['rocket']['radius']},\n"
        text += f"    rocket_radius={self.parameters['rocket']['radius']},\n"
        text += (
            f"    name='{self.parameters['nosecones']['length']}',\n"  # TODO: fix this
        )
        text += ")\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Nosecone created.")
        return nb

    def build_fins(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # add a markdown cell
        text = "### Fins\n"
        text += "As rocketpy allows for multiple fins sets, we will create a "
        text += "dictionary with all the fins sets and then add them to the rocket\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        fin_counter = 0  # count the number of fins
        # trapezoidal fins
        # add a code cell
        if len(self.parameters["trapezoidal_fins"]) > 0:
            self.trapezoidal_fins_check = True
            fin_counter = len(self.parameters["trapezoidal_fins"])
            text = "trapezoidal_fins = {}\n"
            nb["cells"].append(nbf.v4.new_code_cell(text))
            for i in range(len(self.parameters["trapezoidal_fins"])):

                trapezoidal_fins_i = self.parameters["trapezoidal_fins"][str(i)]

                number = trapezoidal_fins_i["number"]
                tip_chord = trapezoidal_fins_i["tip_chord"]
                root_chord = trapezoidal_fins_i["root_chord"]
                span = trapezoidal_fins_i["span"]
                cant_angle = trapezoidal_fins_i["cant_angle"]
                sweep_length = trapezoidal_fins_i["sweep_length"]
                sweep_angle = trapezoidal_fins_i["sweep_angle"]
                radius = self.parameters["rocket"]["radius"]
                name = trapezoidal_fins_i["name"]

                text = f"trapezoidal_fins[{i}] = TrapezoidalFins(\n"
                text += f"    n={number},\n"
                text += f"    root_chord={root_chord},\n"
                text += f"    tip_chord={tip_chord},\n"
                text += f"    span={span},\n"
                text += f"    cant_angle={cant_angle},\n"
                text += f"    sweep_length= {sweep_length},\n"
                text += f"    sweep_angle= {sweep_angle},\n"
                text += f"    rocket_radius={radius},\n"
                text += f"    name='{name}',\n"
                text += ")\n\n"
                nb["cells"].append(nbf.v4.new_code_cell(text))
            logger.info("[NOTEBOOK BUILDER] Trapezoidal fins created.")
        else:
            pass
        # elliptical fins
        # add a code cell
        if len(self.parameters["elliptical_fins"]) > 0:
            self.elliptical_fins_check = True
            fin_counter += len(self.parameters["elliptical_fins"])
            text = "elliptical_fins = {}\n"
            nb["cells"].append(nbf.v4.new_code_cell(text))
            for i in range(len(self.parameters["elliptical_fins"])):

                elliptical_fins_i = self.parameters["elliptical_fins"][str(i)]

                number = elliptical_fins_i["number"]
                root_chord = elliptical_fins_i["root_chord"]
                span = elliptical_fins_i["span"]
                rocket_radius = self.parameters["rocket"]["radius"]
                cant_angle = elliptical_fins_i["cant_angle"]
                name = elliptical_fins_i["name"]

                text = f"elliptical_fins[{i}] = EllipticalFins(\n"
                text += f"    n={number},\n"
                text += f"    root_chord={root_chord},\n"
                text += f"    span={span},\n"
                text += f"    rocket_radius={rocket_radius},\n"
                text += f"    cant_angle={cant_angle},\n"
                text += f"    name='{name}',\n"
                text += ")\n\n"
                nb["cells"].append(nbf.v4.new_code_cell(text))
            logger.info("[NOTEBOOK BUILDER] Elliptical fins created.")
        else:
            pass
        # free form fins
        # checking if fins were added
        try:
            assert fin_counter > 0
            logger.info(
                "[NOTEBOOK BUILDER] %s fins were added to the rocket.", fin_counter
            )
        except AssertionError:
            text = "No fins were added to the rocket. Please add at least one."
            nb["cells"].append(nbf.v4.new_code_cell(text))
            logger.warning("No fins were added to the rocket. Please add at least one.")
            raise Warning("No fins were added to the rocket. Please add at least one.")
        return nb

    def build_tails(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # add a markdown cell
        text = "### Transitions (Tails)\n"
        text += "As rocketpy allows for multiple tails, we will create a "
        text += "dictionary with all the tails and then add them to the rocket\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        # add a code cell
        text = "tails = {}\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))
        for i in range(len(self.parameters["tails"])):

            tail_i = self.parameters["tails"][str(i)]

            top_radius = tail_i["top_radius"]
            bottom_radius = tail_i["bottom_radius"]
            length = tail_i["length"]
            rocket_radius = self.parameters["rocket"]["radius"]
            name = tail_i["name"]

            text = f"tails[{i}] = Tail(\n"
            text += f"    top_radius={top_radius},\n"
            text += f"    bottom_radius={bottom_radius},\n"
            text += f"    length={length},\n"
            text += f"    rocket_radius={rocket_radius},\n"
            text += f"    name='{name}',\n"
            text += ")\n"
            nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Tails created.")
        return nb

    def build_rail_buttons(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        logger.info("rail buttons not implemented yet")
        return nb

    def build_flight(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        """Generates a section defining the flight and returns the notebook."""
        # add a markdown cell
        text = "## Flight\n"
        text += "We will now create the flight simulation. Let's go!\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        # add a code cell
        text = "flight = Flight(\n"
        text += "    rocket=rocket,\n"
        text += "    environment=env,\n"
        text += f"    rail_length={self.parameters['flight']['rail_length']},\n"
        text += f"    inclination={self.parameters['flight']['inclination']},\n"
        text += f"    heading={self.parameters['flight']['heading']},\n"
        text += "    terminate_on_apogee=False,\n"
        text += "    max_time=600,\n"
        text += ")"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        # add a code cell
        text = "flight.all_info()\n"
        nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Flight section created.")
        return nb

    def build_compare_results(self, nb: nbf.v4.new_notebook) -> nbf.v4.new_notebook:
        # pylint: disable=line-too-long
        # add a markdown cell
        text = "## Compare Results\n"
        text += "We will now compare the results of the simulation with the "
        text += "parameters used to create it. Let's go!\n"
        nb["cells"].append(nbf.v4.new_markdown_cell(text))

        # add a code cell
        text = "### OpenRocket vs RocketPy Parameters\n"
        time_to_apogee = self.parameters["stored_results"]["time_to_apogee"]
        # time to apogee
        text += f"time_to_apogee_ork = {time_to_apogee}\n"
        text += "time_to_apogee_rpy = flight.apogee_time\n"
        text += r'print(f"Time to apogee (OpenRocket): {time_to_apogee_ork:.3f} s")'
        text += "\n"
        text += r'print(f"Time to apogee (RocketPy):   {time_to_apogee_rpy:.3f} s")'
        text += "\n"
        text += "apogee_difference = time_to_apogee_rpy - time_to_apogee_ork"
        text += "\n"
        text += "error = abs((apogee_difference)/time_to_apogee_rpy)*100"
        text += "\n"
        text += r'print(f"Time to apogee difference:   {error:.3f} %")'
        text += "\n\n"

        # Flight time
        flight_time_ork = self.parameters["stored_results"]["flight_time"]
        text += f"flight_time_ork = {flight_time_ork}\n"
        text += "flight_time_rpy = flight.t_final\n"
        text += r'print(f"Flight time (OpenRocket): {flight_time_ork:.3f} s")'
        text += "\n"
        text += r'print(f"Flight time (RocketPy):   {flight_time_rpy:.3f} s")'
        text += "\n"
        text += "flight_time_difference = flight_time_rpy - flight_time_ork\n"
        text += (
            "error_flight_time = abs((flight_time_difference)/flight_time_rpy)*100\n"
        )
        text += r'print(f"Flight time difference:   {error_flight_time:.3f} %")'
        text += "\n\n"

        # Ground hit velocity
        ground_hit_velocity_ork = self.parameters["stored_results"][
            "ground_hit_velocity"
        ]
        text += f"ground_hit_velocity_ork = {ground_hit_velocity_ork}\n"
        text += "ground_hit_velocity_rpy = flight.impact_velocity\n"
        text += r'print(f"Ground hit velocity (OpenRocket): {ground_hit_velocity_ork:.3f} m/s")'
        text += "\n"
        text += r'print(f"Ground hit velocity (RocketPy):   {ground_hit_velocity_rpy:.3f} m/s")'
        text += "\n"
        text += "ground_hit_velocity_difference = ground_hit_velocity_rpy - ground_hit_velocity_ork\n"
        text += "error_ground_hit_velocity = abs((ground_hit_velocity_difference)/ground_hit_velocity_rpy)*100\n"
        text += r'print(f"Ground hit velocity difference:   {error_ground_hit_velocity:.3f} %")'
        text += "\n\n"

        # Launch rod velocity
        launch_rod_velocity_ork = self.parameters["stored_results"][
            "launch_rod_velocity"
        ]
        text += f"launch_rod_velocity_ork = {launch_rod_velocity_ork}\n"
        text += "launch_rod_velocity_rpy = flight.out_of_rail_velocity\n"
        text += r'print(f"Launch rod velocity (OpenRocket): {launch_rod_velocity_ork:.3f} m/s")'
        text += "\n"
        text += r'print(f"Launch rod velocity (RocketPy):   {launch_rod_velocity_rpy:.3f} m/s")'
        text += "\n"
        text += "launch_rod_velocity_difference = launch_rod_velocity_rpy - launch_rod_velocity_ork\n"
        text += "error_launch_rod_velocity = abs((launch_rod_velocity_difference)/launch_rod_velocity_rpy)*100\n"
        text += r'print(f"Launch rod velocity difference:   {error_launch_rod_velocity:.3f} %")'
        text += "\n\n"

        # Max acceleration
        max_acceleration_ork = self.parameters["stored_results"]["max_acceleration"]
        text += f"max_acceleration_ork = {max_acceleration_ork}\n"
        text += "max_acceleration_rpy = flight.max_acceleration\n"
        text += (
            r'print(f"Max acceleration (OpenRocket): {max_acceleration_ork:.3f} m/s²")'
        )
        text += "\n"
        text += (
            r'print(f"Max acceleration (RocketPy):   {max_acceleration_rpy:.3f} m/s²")'
        )
        text += "\n"
        text += "max_acceleration_difference = max_acceleration_rpy - max_acceleration_ork\n"
        text += "error_max_acceleration = abs((max_acceleration_difference)/max_acceleration_rpy)*100\n"
        text += (
            r'print(f"Max acceleration difference:   {error_max_acceleration:.3f} %")'
        )
        text += "\n\n"

        # Max altitude
        max_altitude_ork = self.parameters["stored_results"]["max_altitude"]
        text += f"max_altitude_ork = {max_altitude_ork}\n"
        text += "max_altitude_rpy = flight.apogee - flight.env.elevation\n"
        text += r'print(f"Max altitude (OpenRocket): {max_altitude_ork:.3f} m")'
        text += "\n"
        text += r'print(f"Max altitude (RocketPy):   {max_altitude_rpy:.3f} m")'
        text += "\n"
        text += "max_altitude_difference = max_altitude_rpy - max_altitude_ork\n"
        text += (
            "error_max_altitude = abs((max_altitude_difference)/max_altitude_rpy)*100\n"
        )
        text += r'print(f"Max altitude difference:   {error_max_altitude:.3f} %")'
        text += "\n\n"

        # Max Mach
        max_mach_ork = self.parameters["stored_results"]["max_mach"]
        text += f"max_mach_ork = {max_mach_ork}\n"
        text += "max_mach_rpy = flight.max_mach_number \n"
        text += r'print(f"Max Mach (OpenRocket): {max_mach_ork:.3f}")'
        text += "\n"
        text += r'print(f"Max Mach (RocketPy):   {max_mach_rpy:.3f}")'
        text += "\n"
        text += "max_mach_difference = max_mach_rpy - max_mach_ork\n"
        text += "error_max_mach = abs((max_mach_difference)/max_mach_rpy)*100\n"
        text += r'print(f"Max Mach difference:   {error_max_mach:.3f} %")'
        text += "\n\n"

        # Max velocity
        max_velocity_ork = self.parameters["stored_results"]["max_velocity"]
        text += f"max_velocity_ork = {max_velocity_ork}\n"
        text += "max_velocity_rpy = flight.max_speed\n"
        text += r'print(f"Max velocity (OpenRocket): {max_velocity_ork:.3f} m/s")'
        text += "\n"
        text += r'print(f"Max velocity (RocketPy):   {max_velocity_rpy:.3f} m/s")'
        text += "\n"
        text += "max_velocity_difference = max_velocity_rpy - max_velocity_ork\n"
        text += (
            "error_max_velocity = abs((max_velocity_difference)/max_velocity_rpy)*100\n"
        )
        text += r'print(f"Max velocity difference:   {error_max_velocity:.3f} %")'
        text += "\n\n"

        # Max thrust
        max_thrust_ork = self.parameters["stored_results"]["max_thrust"]
        text += f"max_thrust_ork = {max_thrust_ork}\n"
        text += "max_thrust_rpy = flight.rocket.motor.thrust.max\n"
        text += r'print(f"Max thrust (OpenRocket): {max_thrust_ork:.3f} N")'
        text += "\n"
        text += r'print(f"Max thrust (RocketPy):   {max_thrust_rpy:.3f} N")'
        text += "\n"
        text += "max_thrust_difference = max_thrust_rpy - max_thrust_ork\n"
        text += "error_max_thrust = abs((max_thrust_difference)/max_thrust_rpy)*100\n"
        text += r'print(f"Max thrust difference:   {error_max_thrust:.3f} %")'
        text += "\n\n"

        # # Burnout stability margin
        burnout_stability_margin_ork = self.parameters["stored_results"][
            "burnout_stability_margin"
        ]
        text += f"burnout_stability_margin_ork = {burnout_stability_margin_ork}\n"
        text += "burnout_stability_margin_rpy = flight.stability_margin(flight.rocket.motor.burn_out_time)\n"
        text += r'print(f"Burnout stability margin (OpenRocket): {burnout_stability_margin_ork:.3f}")'
        text += "\n"
        text += r'print(f"Burnout stability margin (RocketPy):   {burnout_stability_margin_rpy:.3f}")'
        text += "\n"
        text += "burnout_stability_margin_difference = burnout_stability_margin_rpy - burnout_stability_margin_ork\n"
        text += "error_burnout_stability_margin = abs((burnout_stability_margin_difference)/burnout_stability_margin_rpy)*100\n"
        text += r'print(f"Burnout stability margin difference:   {error_burnout_stability_margin:.3f} %")'
        text += "\n\n"

        # # Max stability margin
        max_stability_margin_ork = self.parameters["stored_results"][
            "max_stability_margin"
        ]
        text += f"max_stability_margin_ork = {max_stability_margin_ork}\n"
        text += "max_stability_margin_rpy = flight.max_stability_margin\n"
        text += r'print(f"Max stability margin (OpenRocket): {max_stability_margin_ork:.3f}")'
        text += "\n"
        text += r'print(f"Max stability margin (RocketPy):   {max_stability_margin_rpy:.3f}")'
        text += "\n"
        text += "max_stability_margin_difference = max_stability_margin_rpy - max_stability_margin_ork\n"
        text += "error_max_stability_margin = abs((max_stability_margin_difference)/max_stability_margin_rpy)*100\n"
        text += r'print(f"Max stability margin difference:   {error_max_stability_margin:.3f} %")'
        text += "\n\n"

        # Min stability margin
        min_stability_margin_ork = self.parameters["stored_results"][
            "min_stability_margin"
        ]
        text += f"min_stability_margin_ork = {min_stability_margin_ork}\n"
        text += "min_stability_margin_rpy = flight.min_stability_margin\n"
        text += r'print(f"Min stability margin (OpenRocket): {min_stability_margin_ork:.3f}")'
        text += "\n"
        text += r'print(f"Min stability margin (RocketPy):   {min_stability_margin_rpy:.3f}")'
        text += "\n"
        text += "min_stability_margin_difference = min_stability_margin_rpy - min_stability_margin_ork\n"
        text += "error_min_stability_margin = abs((min_stability_margin_difference)/min_stability_margin_rpy)*100\n"
        text += r'print(f"Min stability margin difference:   {error_min_stability_margin:.3f} %")'
        text += "\n\n"

        nb["cells"].append(nbf.v4.new_code_cell(text))

        logger.info("[NOTEBOOK BUILDER] Compare Results section created.")
        return nb

    def save_notebook(self, nb: nbf.v4.new_notebook, destination: str) -> None:
        """Writes the .ipynb file to the destination folder. Also applies black
        formatting to the file to improve readability."""
        out_file = os.path.join(destination, "simulation.ipynb")

        nbf.write(nb, out_file)
        logger.info("[NOTEBOOK BUILDER] Notebook saved to '%s'", out_file)

        # apply black formatting after saving (requires black[jupyter])
        os.system(f"black {out_file}")
        logger.info("[NOTEBOOK BUILDER] Black formatting applied to the final notebook")



import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_trapezoidal_fins(bs, elements):
    """Search for trapezoidal fins in the bs and return the settings as a dict.
    It is flexible in the sense that it can handle multiple trapezoidal fin sets.

    Parameters
    ----------
    bs : BeautifulSoup
        The BeautifulSoup object of the open rocket file.
    elements : dict
        Dictionary with the settings for the elements of the rocket.

    Returns
    -------
    settings : dict
        Dictionary with the settings for the trapezoidal fins. The keys are
        integers and the values are dicts containing the settings for each
        trapezoidal fin set. The keys of the trapezoidal fin set dicts are:
        "name", "number", "root_chord", "tip_chord", "span", "position",
        "sweep_length", "sweep_angle", "cant_angle", "section".
    """
    settings = {}
    fins = bs.findAll("trapezoidfinset")
    logger.info("A total of %d trapezoidal fin sets were detected", len(fins))

    if len(fins) == 0:
        logger.info("No trapezoidal fins were detected, returning empty dictionary")
        return settings

    for idx, fin in enumerate(fins):
        logger.info(
            "Starting collecting the settings for the trapezoidal fin set number '%d'",
            idx,
        )
        label = fin.find("name").text
        try:

            def get_element_by_name(name):
                for element in elements.values():
                    if element["name"] == name:
                        return element
                return None

            element = get_element_by_name(label)
            logger.info("Found the element '%s' in the elements dictionary.", label)
        except KeyError:
            message = (
                f"Couldn't find the element '{label}' in the elements dictionary. It is"
                " possible that the process_elements_position() function got an error."
            )
            logger.error(message)
            raise KeyError(message)

        n_fin = int(fin.find("fincount").text)
        logger.info("Number of fins retrieved: %d", n_fin)

        root_chord = float(fin.find("rootchord").text)
        logger.info("Root chord retrieved: %f", root_chord)

        tip_chord = float(fin.find("tipchord").text)
        logger.info("Tip chord retrieved: %f", tip_chord)

        span = float(fin.find("height").text)
        logger.info("Span retrieved: %f", span)

        sweep_length = (
            float(fin.find("sweeplength").text) if fin.find("sweeplength") else None
        )
        sweep_angle = (
            float(fin.find("sweepangle").text) if fin.find("sweepangle") else None
        )
        logger.info(
            "Sweep length and angle retrieved: %s, %s", sweep_length, sweep_angle
        )

        cant_angle = float(fin.find("cant").text)
        logger.info("Cant angle retrieved: %f", cant_angle)

        section = fin.find("crosssection").text
        logger.info("Crosssection format retrieved")

        fin_settings = {
            "name": label,
            "number": n_fin,
            "root_chord": root_chord,
            "tip_chord": tip_chord,
            "span": span,
            "position": element["position"],
            "sweep_length": sweep_length,
            "sweep_angle": sweep_angle,
            "cant_angle": cant_angle,
            "section": section,
        }

        settings[idx] = fin_settings

        logger.info(
            "Trapezoidal fin set number '%d' was defined:\n%s",
            idx,
            _dict_to_string(fin_settings, indent=23),
        )

    logger.info("Finished collecting all the trapezoidal fins.")
    return settings


def search_elliptical_fins(bs, elements):
    """Search for elliptical fins in the bs and return the settings as a dict.
    It is flexible in the sense that it can handle multiple elliptical fin sets.

    Parameters
    ----------
    bs : BeautifulSoup
        The BeautifulSoup object of the open rocket file.
    elements : dict
        Dictionary with the settings for the elements of the rocket.

    Returns
    -------
    settings : dict
        Dictionary with the settings for the elliptical fins. The keys are
        integers and the values are dicts containing the settings for each
        elliptical fin set. The keys of the elliptical fin set dicts are:
        "name", "number", "root_chord", "span", "position", "cant_angle",
        "section".
    """
    settings = {}
    fins = bs.findAll("ellipticalfinset")
    logger.info("A total of %d elliptical fin sets were detected", len(fins))

    if len(fins) == 0:
        logger.info(
            "Since no elliptical fins were detected, returning empty dictionary"
        )
        return settings

    for idx, fin in enumerate(fins):
        logger.info(
            "Starting collecting the settings for the elliptical fin set number '%d'",
            idx,
        )
        label = fin.find("name").text
        try:

            def get_element_by_name(name):
                for element in elements.values():
                    if element["name"] == name:
                        return element
                return None

            element = get_element_by_name(label)
            logger.info("Found the element '%s' in the elements dictionary.", label)
        except KeyError:
            message = (
                f"Couldn't find the element '{label}' in the elements dictionary. It is"
                " possible that the process_elements_position() function got an error."
            )
            logger.error(message)
            raise KeyError(message)

        n_fin = int(fin.find("fincount").text)
        logger.info("Number of fins retrieved: %d", n_fin)

        root_chord = float(fin.find("rootchord").text)
        logger.info("Root chord retrieved: %f", root_chord)

        span = float(fin.find("height").text)
        logger.info("Span retrieved: %f", span)

        cant_angle = float(fin.find("cant").text)
        logger.info("Cant angle retrieved: %f", cant_angle)

        section = fin.find("crosssection").text
        logger.info("Crosssection format retrieved")

        fin_settings = {
            "name": label,
            "number": n_fin,
            "root_chord": root_chord,
            "span": span,
            "position": element["position"],
            "cant_angle": cant_angle,
            "section": section,
        }

        settings[idx] = fin_settings

        logger.info(
            "Elliptical fin set number '%d' was defined:\n%s",
            idx,
            _dict_to_string(fin_settings, indent=23),
        )

    logger.info("Finished collecting all the elliptical fins.")
    return settings


def search_free_form_fins(bs, elements):  # pylint: disable=unused-argument
    return {}


# TODO: support for tubefinset (low priority)



import logging
from pathlib import Path

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_id_info(bs, filepath):
    """Searches for the identification of the .ork file

    Parameters
    ----------
    bs : BeautifulSoup
        BeautifulSoup object of the .ork file.
    filepath : str
        Path to the .ork file.

    Returns
    -------
    dictionary
        Dictionary with the identification information of the .ork file. The
        keys are: "rocket_name", "comment", "designer", "ork_version" and
        "filepath".
    """
    settings = {}
    settings["rocket_name"] = bs.find("rocket").find("name").text
    logger.info("Collected the rocket name: '%s'", settings["rocket_name"])

    try:
        settings["comment"] = bs.find("rocket").find("comment").text.replace("\n", "")
        logger.info("Collected the comment saved in the file: %s", settings["comment"])
    except AttributeError:
        logger.warning("No auxiliary comment was found in the file.")
        settings["comment"] = None
    try:
        settings["designer"] = bs.find("rocket").find("designer").text
        logger.info("Collected the designer name: %s", settings["designer"])
    except AttributeError:
        logger.warning("No designer name was found in the file.")
        settings["designer"] = None
    # settings["ork_version"] = bs.attrs["creator"]
    settings["filepath"] = Path(filepath).as_posix()

    logger.info(
        "Identification information extracted.\n %s",
        _dict_to_string(settings, indent=23),
    )
    return settings



import json
import logging
import os
from pathlib import Path

import click
import orhelper

from ._helpers import extract_ork_from_zip, parse_ork_file
from .nb_builder import NotebookBuilder
from .ork_extractor import ork_extractor

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y.%b.%d %H:%M:%S",
    filename="serializer.log",
    filemode="w",
)
logger = logging.getLogger(__name__)

console = logging.StreamHandler()
console.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
console.setFormatter(formatter)
logger.addHandler(console)


@click.group()
@click.version_option()
def cli():
    """RocketSerializer.
    This library has as objective to convert .ork files into parameters.json, so
    that they can be used in rocketpy simulations. It also provides the option
    to convert the parameters.json file into a .ipynb file, so that the user can
    run the simulation using Jupyter Notebooks.

    Examples
    --------
    To easily use the library, you can use the command line interface. For
    example, to generate a .json file from a .ork file, you can use the
    following command on your terminal:
    >>> ork2json("rocket.ork", "rocket", "motor.eng")

    If you want to use the library with Python, you can import the library and
    use the functions directly. For example, to generate a .json file from a
    .ork file, you can use the following code:

    >>> from rocketserializer import ork2json
    >>> ork2json([ "--filepath", "rocket.ork", "--eng", "motor.eng"])

    If you want to convert a .ork file to a Jupyter Notebook, you can use the
    following command on your terminal:

    >>> ork2notebook("rocket.ork", "rocket", "motor.eng")
    """


@cli.command("ork2json")
@click.option(
    "--filepath", type=click.Path(), required=True, help="The path to the .ork file."
)
@click.option(
    "--output", type=click.Path(), required=False, help="The path to the output folder."
)
@click.option(
    "--ork_jar",
    type=click.Path(),
    default=None,
    required=False,
    help="The path to the OpenRocket .jar file.",
)
@click.option("--encoding", type=str, default="utf-8", required=False)
@click.option("--verbose", type=bool, default=False, required=False)
def ork2json(filepath, output=None, ork_jar=None, encoding="utf-8", verbose=False):
    """Generates a .json file from the .ork file.
    The .json file will be generated in the output folder using the information
    of the .ork file. It is possible to specify the .eng file to extract the
    thrust curve from it. If the .eng file is not specified, the thrust curve
    will be extracted from the .ork file.

    Parameters
    ----------
    filepath : str
        The path to the .ork file.
    output : str
        The path to the output folder.
    ork_jar : str, optional
        The path to the OpenRocket .jar file. If unspecified, the .jar file
        will be searched in the current directory.
    encoding : str, optional
        The encoding of the .json file. Default is 'utf-8'.
    verbose : bool, optional
        If True, the log level will be set to DEBUG. Default is False.

    Raises
    ------
    ValueError
        In case the .ork file does not contain the simulation data.
    ValueError
        In case the .ork file is not in English.
    """
    log_level = logging.DEBUG if verbose else logging.WARNING
    logger.setLevel(log_level)

    filepath = Path(filepath)

    if not filepath.exists():
        error = (
            "[ork2json] The .ork file or zip archive does not exist. "
            "Please specify a valid path."
        )
        logger.error(error)
        raise FileNotFoundError(error)

    if filepath.suffix.lower() == ".ork":
        extract_dir = filepath.parent
        filepath = extract_ork_from_zip(filepath, extract_dir)
        logger.info("[ork2json] Extracted .ork file to: %s", filepath.as_posix())

    bs, datapoints = parse_ork_file(filepath)

    if len(datapoints) == 0:
        error_msg = (
            "[ork2json] The file must contain the simulation data.\n"
            + "Open the .ork file and run the simulation first."
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    data_labels = bs.find("databranch").attrs["types"].split(",")
    if "CG location" not in data_labels:
        message = (
            "[ork2json] The file must contain the simulation data.\n"
            + "Open the .ork file and run the simulation first."
        )
        logger.error(message)
        raise ValueError(message)

    if not ork_jar:
        # get any .jar file in the current directory that starts with "OpenRocket"
        ork_jar = [
            f for f in os.listdir() if f.startswith("OpenRocket") and f.endswith(".jar")
        ]
        if len(ork_jar) == 0:
            raise ValueError(
                "[ork2json] It was not possible to find the OpenRocket .jar file in "
                "the current directory. Please specify the path to the .jar file."
            )
        ork_jar = ork_jar[0]
        logger.info(
            "[ork2json] Found OpenRocket .jar file: '%s'", Path(ork_jar).as_posix()
        )

    if not output:
        # get the same folder as the .ork file
        output = os.path.dirname(filepath)
        logger.warning(
            "[ork2json] Output folder not specified. Using '%s' instead.",
            Path(output).as_posix(),
        )

    # orhelper options are: OFF, ERROR, WARN, INFO, DEBUG, TRACE and ALL
    # log_level = "OFF" if verbose else "OFF"
    # TODO: even if the log level is set to OFF, the orhelper still prints msgs

    with orhelper.OpenRocketInstance(ork_jar, log_level="OFF") as instance:
        # create the output folder if it does not exist
        if os.path.exists(output) is False:
            os.mkdir(output)

        orh = orhelper.Helper(instance)
        ork = orh.load_doc(str(filepath))

        settings = ork_extractor(
            bs=bs,
            filepath=str(filepath),
            output_folder=output,
            ork=ork,
        )

        with open(
            os.path.join(output, "parameters.json"), "w", encoding=encoding
        ) as convert_file:
            convert_file.write(
                json.dumps(settings, indent=4, sort_keys=True, ensure_ascii=False)
            )
            logger.info(
                "[ork2json] The file 'parameters.json' was saved to: '%s'",
                Path(output).as_posix(),
            )
            logger.info(
                "[ork2json] Operation completed successfully. You can now use "
                "the 'parameters.json' file to run a simulation."
            )


@cli.command("ork2notebook")
@click.option("--filepath", type=str, required=True)
@click.option("--output", type=str, required=False)
@click.option("--ork_jar", type=str, default=None, required=False)
@click.option("--encoding", type=str, default="utf-8", required=False)
@click.option("--verbose", type=bool, default=False, required=False)
def ork2notebook(filepath, output, ork_jar=None, encoding="utf-8", verbose=False):
    """Generates a .ipynb file from the .ork file.

    Notes
    -----
    Under the hood, this function uses the `ork2json` function to generate the
    parameters.json file and then uses the `NotebookBuilder` class to generate
    the .ipynb file.
    """
    ork2json(
        [
            "--filepath",
            filepath,
            "--output",
            output,
            "--ork_jar",
            ork_jar,
            "--encoding",
            encoding,
            "--verbose",
            verbose,
        ],
        standalone_mode=True,
    )
    instance = NotebookBuilder(parameters_json=os.path.join(output, "parameters.json"))
    instance.build(destination=output)



import logging
import os
from pathlib import Path

import numpy as np

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_motor(bs, datapoints, data_labels):
    """Search for the motor properties in the .ork file. The only property that
    is not included is the thrust curve, which is generated in the
    generate_thrust_curve function. Only rocketpy.SolidMotor class would be able
    to use the information from this function to create a motor object.

    Parameters
    ----------
    bs : bs4.BeautifulSoup
        The BeautifulSoup object of the .ork file.
    datapoints : list
        The datapoints from the .ork file.
    data_labels : list
        The labels of the datapoints.
    time_vector : list
        The time vector of the simulation.

    Returns
    -------
    settings : dict
        Dictionary with the motor properties. The keys are: "burn_time",
        "grain_density", "grain_initial_inner_radius", "grain_outer_radius",
        "grain_initial_height", "nozzle_radius", "throat_radius", "dry_mass",
        "dry_inertia", "center_of_dry_mass", "grains_center_of_mass_position",
        "grain_number", "grain_separation", "nozzle_position" and
        "coordinate_system_orientation".
    """
    settings = {}

    # retrieve motor geometry
    motor_length = float(bs.find("motormount").find("length").text)
    motor_radius = float(bs.find("motormount").find("diameter").text) / 2
    logger.info("Collected motor geometry: motor length and motor radius.")

    # get motor mass properties
    total_propellant_mass, motor_dry_mass, _ = __get_motor_mass(datapoints, data_labels)
    motor_dry_mass = 0  # If NOTE: dry inertia is 0, this should ALWAYS be 0 too.
    center_of_dry_mass = 0
    dry_inertia = (0, 0, 0)  # impossible to retrieve from .ork file

    # define grains properties
    grain_number = 1
    grain_separation = 0
    grain_initial_inner_radius = motor_radius / 2
    grain_outer_radius = motor_radius
    grain_initial_height = motor_length
    grain_volume = (
        np.pi
        * (grain_outer_radius**2 - grain_initial_inner_radius**2)
        * grain_initial_height
    ) / grain_number
    grain_density = total_propellant_mass / (grain_volume * grain_number)
    grains_center_of_mass_position = 0
    logger.info("Calculated motor mass properties.")

    # get nozzle properties (impossible to retrieve from .ork file)
    throat_radius = 1.0 * grain_initial_inner_radius
    nozzle_radius = 1.5 * grain_initial_inner_radius
    nozzle_position = -motor_length / 2

    # set other motor properties
    coordinate_system_orientation = "nozzle_to_combustion_chamber"

    settings = {
        "grain_density": grain_density,
        "grain_initial_inner_radius": grain_initial_inner_radius,
        "grain_outer_radius": motor_radius,
        "grain_initial_height": motor_length,
        "nozzle_radius": nozzle_radius,
        "throat_radius": throat_radius,
        "dry_mass": motor_dry_mass,
        "dry_inertia": dry_inertia,
        "center_of_dry_mass_position": center_of_dry_mass,
        "grains_center_of_mass_position": grains_center_of_mass_position,
        "grain_number": grain_number,
        "grain_separation": grain_separation,
        "nozzle_position": nozzle_position,
        "coordinate_system_orientation": coordinate_system_orientation,
    }
    logger.info(
        "Successfully configured the motor.\n %s", _dict_to_string(settings, indent=23)
    )
    return settings


def generate_thrust_curve(folder_path, datapoints, data_labels, time_vector):
    """Generate the thrust curve from the .ork file.

    Parameters
    ----------
    folder_path : str
        The path to the folder where the thrust curve should be saved. This
        needs to be a folder that already exists.
    datapoints : list
        List of datapoints from the .ork file.
    data_labels : list
        List of labels for the datapoints.
    time_vector : list
        The time vector of the simulation.

    Returns
    -------
    source_name : str
        The path to the thrust curve.
    """
    thrust = [
        float(datapoint.text.split(",")[data_labels.index("Thrust")])
        for datapoint in datapoints
    ]
    logger.info("Collected thrust vector")

    # convert to numpy array
    thrust = np.array([time_vector, thrust]).T

    # sort by time
    thrust = thrust[thrust[:, 0].argsort()]
    logger.info("The thrust points were sorted to be in ascending order")

    # clip the thrust curve to remove negative values
    thrust[thrust[:, 1] < 0, 1] = 0
    logger.info("Successfully clipped the thrust curve to remove negative values")

    # remove any items with thrust lower than 0.0001 N
    thrust = thrust[thrust[:, 1] > 0.0001, :]
    logger.info("Successfully created the thrust curve")

    # save to a csv file
    source_name = os.path.join(folder_path, "thrust_source.csv")
    np.savetxt(source_name, thrust, delimiter=",", fmt="%1.5f")

    logger.info(
        "Successfully saved the thrust curve to: '%s'", Path(source_name).as_posix()
    )
    return source_name


def __get_motor_mass(datapoints, data_labels):
    """Get the motor mass from the .ork file.

    Parameters
    ----------
    datapoints : list
        List of datapoints from the .ork file.
    data_labels : list
        List of labels for the datapoints.

    Returns
    -------
    total_propellant_mass : float
        Total propellant mass, in kg.
    motor_dry_mass : float
        Motor dry mass, in kg. It is the minimum value of the motor mass. This
        corresponds to the mass of the motor without the propellant.
    burnout_position : int
        The index of the burnout position in the datapoints list.
    """
    if "Propellant mass" in data_labels:
        prop_mass_vector = [
            float(datapoint.text.split(",")[data_labels.index("Propellant mass")])
            for datapoint in datapoints
        ]
        motor_dry_mass = min(prop_mass_vector)
        logger.info("The motor dry mass is %.3f kg.", motor_dry_mass)
    elif "Motor mass" in data_labels:
        motor_mass = np.array(
            [
                float(datapoint.text.split(",")[data_labels.index("Motor mass")])
                for datapoint in datapoints
            ]
        )
        motor_dry_mass = min(motor_mass)
        prop_mass_vector = motor_mass - motor_dry_mass
        prop_mass_vector = list(prop_mass_vector)
        logger.info("The motor dry mass is %.3f kg.", motor_dry_mass)

    normalize = np.array(prop_mass_vector)
    normalize = normalize - normalize[np.argmin(normalize)]
    prop_mass_vector = list(normalize)
    total_propellant_mass = prop_mass_vector[0]
    burnout_position = np.argwhere(np.array(prop_mass_vector) == 0)[0][0]
    # q: why is it important to normalize the propellant mass vector?
    # a: to ensure that we are using the propellant mass, without the motor mass.
    #    Also, to ensure the final propellant mass is zero.

    logger.info(
        "The total propellant mass is %.3f kg. The motor dry mass is %.3f kg.",
        total_propellant_mass,
        motor_dry_mass,
    )
    return total_propellant_mass, motor_dry_mass, burnout_position



import logging

from ._helpers import _dict_to_string
from .components.drag_curve import save_drag_curve
from .components.environment import search_environment
from .components.fins import search_elliptical_fins, search_trapezoidal_fins
from .components.flight import search_launch_conditions
from .components.id import search_id_info
from .components.motor import __get_motor_mass, generate_thrust_curve, search_motor
from .components.nose_cone import search_nosecone
from .components.open_rocket_wrangler import process_elements_position
from .components.parachute import search_parachutes
from .components.rail_buttons import search_rail_buttons
from .components.rocket import search_rocket
from .components.stored_results import search_stored_results
from .components.transition import search_transitions

logger = logging.getLogger(__name__)


def ork_extractor(bs, filepath, output_folder, ork):
    """Generates the parameters.json file with the parameters for rocketpy

    Parameters
    ----------
    bs : BeautifulSoup
        BeautifulSoup object of the .ork file.
    filepath : str
        Path to the .ork file.
    output_folder : str
        Path to the output folder.
    ork : orhelper
        An object representing the OpenRocket document.
    verbose : bool, optional
        Whether or not to print a message of successful execution, by default
        False.

    Returns
    -------
    dictionary
        Dictionary with the parameters to be used in rocketpy simulations. The
        keys are: "id", "environment", "rocket", "nosecones", "trapezoidal_fins",
        "tails", "parachutes", "rail_buttons", "motors", "flight" and
        "stored_results".
    """
    settings = {}

    # Initialize some important vectors
    datapoints, data_labels, time_vector = __init_vectors(bs)
    logger.info("Initialized data vectors from the ORK file.")

    # Retrieve the motor properties
    motors = search_motor(bs, datapoints, data_labels)
    _, _, burnout_position = __get_motor_mass(datapoints, data_labels)
    logger.info("Motor parameters retrieved.")

    # Get the first set of parameters
    id_info = search_id_info(bs, filepath)
    logger.info("Metadata parameters retrieved.")

    environment = search_environment(bs)
    logger.info("Environment parameters retrieved.")

    rocket, motor_position = search_rocket(
        bs, datapoints, data_labels, burnout_position
    )
    motors["position"] = motor_position
    logger.info("Rocket parameters retrieved.")

    flight = search_launch_conditions(bs)
    logger.info("Flight conditions retrieved.")

    # process different elements of the rocket
    center_of_dry_mass = rocket["center_of_mass_without_propellant"]
    rocket_mass = rocket["mass"]
    rocket_radius = rocket["radius"]

    elements = process_elements_position(
        ork.getRocket(), {}, center_of_dry_mass, rocket_mass, top_position=0
    )
    logger.info("The elements are:\n%s", _dict_to_string(elements, indent=23))

    nosecones = search_nosecone(bs, elements, rocket_radius)
    trapezoidal_fins = search_trapezoidal_fins(bs, elements)
    elliptical_fins = search_elliptical_fins(bs, elements)
    transitions = search_transitions(bs, elements, ork)
    rail_buttons = search_rail_buttons(bs, elements)
    parachutes = search_parachutes(bs)
    stored_results = search_stored_results(
        bs, datapoints, data_labels, time_vector, burnout_position
    )

    # save everything to a dictionary
    settings["id"] = id_info
    settings["environment"] = environment
    settings["rocket"] = rocket
    settings["nosecones"] = nosecones
    settings["trapezoidal_fins"] = trapezoidal_fins
    settings["elliptical_fins"] = elliptical_fins
    settings["tails"] = transitions
    settings["parachutes"] = parachutes
    settings["rail_buttons"] = rail_buttons
    settings["motors"] = motors
    settings["flight"] = flight
    settings["stored_results"] = stored_results

    # get drag curves
    settings["rocket"]["drag_curve"] = save_drag_curve(
        datapoints, data_labels, output_folder
    )
    logger.info("Drag curve generated.")

    # get thrust curve
    thrust_path = generate_thrust_curve(
        output_folder, datapoints, data_labels, time_vector
    )
    settings["motors"]["thrust_source"] = thrust_path
    logger.info("Thrust curve generated.")

    logger.info(
        "Extraction completed. A dictionary with all the parameters was generated."
    )
    logger.info(
        "Dictionary with the parameters:\n%s", _dict_to_string(settings, indent=23)
    )

    return settings


def __init_vectors(bs):
    """Initializes the vectors with the data from the .ork file.

    Parameters
    ----------
    bs : BeautifulSoup
        BeautifulSoup object of the .ork file.

    Returns
    -------
    datapoints : list
        The datapoints.
    data_labels : list
        The names of each data column.
    time_vector : list
        The time vector.
    """
    datapoints = bs.findAll("datapoint")
    data_labels = bs.find("databranch").attrs["types"].split(",")

    time_vector = [float(datapoint.text.split(",")[0]) for datapoint in datapoints]
    start_pos = 0
    final_pos = len(time_vector) - 1

    # Get the start position, the ignition time.
    for idx, position in enumerate(time_vector):
        if position == 0:
            start_pos = idx

    # Filter the datapoints to get only the ones after the ignition.
    datapoints = datapoints[start_pos:final_pos]
    time_vector = time_vector[start_pos:final_pos]
    logger.info("Successfully initialized vectors with %d datapoints", len(datapoints))
    return datapoints, data_labels, time_vector



import logging
import os
from pathlib import Path

import numpy as np

logger = logging.getLogger(__name__)


def save_drag_curve(datapoints, data_labels, output_folder):
    """Extracts the drag curve from the data and saves it to a csv file.

    Parameters
    ----------
    datapoints : list
        The datapoints.
    data_labels : list of str
        The labels of the data.
    path : str
        Path to the folder where the drag curve should be saved. This needs to
        be a folder that already exists.

    Returns
    -------
    path : str
        The path to the drag curve.
    """
    # Remove the data after apogee
    altitude_vector = [
        float(datapoint.text.split(",")[data_labels.index("Altitude")])
        for datapoint in datapoints
    ]
    logger.info("Collected altitude vector")

    apogee_index = np.argmax(altitude_vector)
    datapoints = datapoints[:apogee_index]
    logger.info("Removed data after apogee")

    # Extract the drag coefficient and Mach number
    cd = [
        float(datapoint.text.split(",")[data_labels.index("Axial drag coefficient")])
        for datapoint in datapoints
    ]
    mach = [
        float(datapoint.text.split(",")[data_labels.index("Mach number")])
        for datapoint in datapoints
    ]
    logger.info("Collected drag coefficient and Mach number")

    # Convert to numpy array
    cd = np.array([mach, cd]).T
    # Remove NaN values to avoid errors
    cd = cd[~np.isnan(cd).any(axis=1), :]
    # Sort by Mach number
    cd = cd[cd[:, 0].argsort()]
    # Remove duplicate Mach numbers
    cd = np.unique(cd, axis=0)
    # Remove values when the drag is lower than 0
    cd = cd[cd[:, 1] > 0, :]
    logger.info("Successfully created the drag curve")

    # Save to csv file
    path = os.path.join(output_folder, "drag_curve.csv")
    np.savetxt(path, cd, delimiter=",", fmt="%.6f")
    logger.info(
        "Successfully saved the drag curve file to: '%s'", Path(path).as_posix()
    )
    return path



import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def search_environment(bs):
    """Searches the launch conditions in the bs object. Returns a dict with the
    settings.

    Parameters
    ----------
    bs : BeautifulSoup
        BeautifulSoup object of the .ork file.

    Returns
    -------
    settings : dict
        A dict containing the settings for the launch conditions. The keys are:
        "latitude", "longitude", "elevation", "wind_average", "wind_turbulence",
        "geodetic_method", "base_temperature" and "base_pressure".
    """
    settings = {}

    latitude = float(bs.find("launchlatitude").text)
    longitude = float(bs.find("launchlongitude").text)
    elevation = float(bs.find("launchaltitude").text)
    wind_average = float(bs.find("windaverage").text)
    wind_turbulence = float(bs.find("windturbulence").text)
    geodetic_method = bs.find("geodeticmethod").text
    logger.info(
        "Collected first environment settings: latitude, "
        + "longitude, elevation, wind_average, wind_turbulence, geodetic_method"
    )
    try:
        base_temperature = float(bs.find("basetemperature").text)
        logger.info(
            "The base temperature was found in the .ork file. It is %f °C.",
            base_temperature,
        )
    except AttributeError:
        logger.warning(
            "No base temperature was found in the .ork file. "
            + "The base temperature will be set to None."
        )
        base_temperature = None
    try:
        base_pressure = float(bs.find("basepressure").text)
        logger.info(
            "The base pressure was found in the .ork file. It is %f Pa.",
            base_pressure,
        )
    except AttributeError:
        logger.warning(
            "No base pressure was found in the .ork file. "
            + "The base pressure will be set to None."
        )
        base_pressure = None
    date = None

    settings = {
        "latitude": latitude,
        "longitude": longitude,
        "elevation": elevation,
        "wind_average": wind_average,
        "wind_turbulence": wind_turbulence,
        "geodetic_method": geodetic_method,
        "base_temperature": base_temperature,
        "base_pressure": base_pressure,
        "date": date,
    }
    logger.info(
        "Successfully extracted all the environment settings.\n %s",
        _dict_to_string(settings, indent=23),
    )
    return settings



import logging

from .._helpers import _dict_to_string

logger = logging.getLogger(__name__)


def is_sub_component(ork):
    """Determines whether the given object `ork` is a sub-component of another
    object in a hierarchical structure.

    Parameters:
    -----------
    ork: net.sf.openrocket.rocketcomponent
        The object to be checked.

    Returns:
    --------
    bool
        True if `ork` is a sub-component (at least two levels below the root
        object), False otherwise.
    """
    level = 0
    root = ork.getRoot()
    while ork != root:
        ork = ork.getParent()
        level += 1
    return level > 1


def parent_is_a_stage(ork) -> bool:
    # Verify if the parent of the object is a Stage or not.
    try:
        parent = ork.getParent()
    except AttributeError:
        return False

    return parent.getClass().getSimpleName() in ["Stage", "AxialStage"]


def calculate_distance_to_the_nose_tip(ork, top_position: float):
    # NOTE: Always considering the nose tip as the reference point, it's the 0.

    if ork.getClass().getSimpleName() in ["Rocket"]:
        return 0

    if is_sub_component(ork) is True:
        try:
            distance_to_nose = top_position + ork.getPositionValue()
        except AttributeError:
            # "object has no attribute 'getPositionValue'"
            if parent_is_a_stage(ork):
                distance_to_nose = top_position
            else:
                distance_to_nose = top_position + ork.getPosition().x
    else:
        distance_to_nose = top_position + ork.getLength()
    return float(distance_to_nose)


def process_elements_position(
    ork, elements, center_of_dry_mass, rocket_mass, top_position=0
):
    if ork.getClass().getSimpleName() in ["Parachute", "MassComponent"]:
        # These classes are irrelevant for our work
        return elements

    element = {
        "type": str(ork.getClass().getSimpleName()),
        "name": str(ork.getName()),
        "length": float(ork.getLength()),
        "position": calculate_distance_to_the_nose_tip(ork, top_position),
    }
    logger.info(
        "Starting to process element '%s' of type '%s'",
        ork.getName(),
        ork.getClass().getSimpleName(),
    )
    unique_id = hash(frozenset(element.items()))
    elements[unique_id] = element
    i = 0

    while True:
        if i >= ork.getChildCount():
            logger.info("No more children for '%s'", ork.getName())
            break

        try:
            child = ork.getChild(i)
            new_elements = process_elements_position(
                child, {}, center_of_dry_mass, rocket_mass, top_position
            )
            elements.update(new_elements)
            logger.info("Child '%s' processed", child.getName())

            top_position += child.getLength()

            i += 1
            logger.info("Moving to the next child")

        except Exception as e:  # pylint: disable=broad-except
            if "IndexOutOfBoundsException" in str(e):
                logger.warning(
                    "Index out of bounds - likely no more children to process."
                )
                break
            else:
                logger.error(
                    "Error while processing the position of the elements: %s",
                    e,
                    exc_info=True,
                )
                logger.info(
                    "The elements are:\n%s",
                    _dict_to_string(elements, indent=23),
                )
                print(e)
                break  # Exit the loop if an unexpected error occurs

    logger.info("Finished processing '%s'", ork.getName())
    return elements
